# ADXL345.py
# Version 1.00, Jan. 30, 2019
# Original from Jonathan Williamson. Ported to MicroPython

from time import sleep
from utils import I2C_com
from machine import Pin, I2C
from math import asin, atan2, sqrt, degrees

class ADXL345:
    EARTH_GRAVITY_MS2   = 9.80665
    SCALE_MULTIPLIER    = 0.004
    
    DATA_FORMAT         = const(0x31)
    BW_RATE             = const(0x2C)
    POWER_CTL           = const(0x2D)
    
    BW_RATE_1600HZ      = const(0x0F)
    BW_RATE_800HZ       = const(0x0E)
    BW_RATE_400HZ       = const(0x0D)
    BW_RATE_200HZ       = const(0x0C)
    BW_RATE_100HZ       = const(0x0B)
    BW_RATE_50HZ        = const(0x0A)
    BW_RATE_25HZ        = const(0x09)
    
    RANGE_2G            = const(0x00)
    RANGE_4G            = const(0x01)
    RANGE_8G            = const(0x02)
    RANGE_16G           = const(0x03)
    
    MEASURE             = const(0x08)
    AXES_DATA           = const(0x32)    
    
    def __init__(self, scl = Pin(22), sda = Pin(21), addr = 0x53):
        self._i2c_com = I2C_com(addr, scl, sda)
        self._setBandwidthRate(BW_RATE_100HZ)
        self._setRange(RANGE_2G)
        self._enableMeasurement()

    def _enableMeasurement(self):
        self._i2c_com.writeReg(POWER_CTL, MEASURE)

    def _setBandwidthRate(self, rate_flag):
        self._i2c_com.writeReg(BW_RATE, rate_flag)

    # set the measurement range for 10-bit readings
    def _setRange(self, range_flag):
        value = self._i2c_com.readReg(DATA_FORMAT)

        value &= ~0x0F;
        value |= range_flag;  
        value |= 0x08;

        self._i2c_com.writeReg(DATA_FORMAT, value)
    
    def getAccel(self):
        bytes = self._i2c_com.readBlock(AXES_DATA, 6)
      
        x = bytes[0] | (bytes[1] << 8)
        if(x & (1 << 16 - 1)):
            x = x - (1<<16)

        y = bytes[2] | (bytes[3] << 8)
        if(y & (1 << 16 - 1)):
            y = y - (1<<16)

        z = bytes[4] | (bytes[5] << 8)
        if(z & (1 << 16 - 1)):
            z = z - (1<<16)

        x = x * ADXL345.SCALE_MULTIPLIER 
        y = y * ADXL345.SCALE_MULTIPLIER
        z = z * ADXL345.SCALE_MULTIPLIER

        x = x * ADXL345.EARTH_GRAVITY_MS2
        y = y * ADXL345.EARTH_GRAVITY_MS2
        z = z * ADXL345.EARTH_GRAVITY_MS2

        ax = round(x, 4)
        ay = round(y, 4)
        az = round(z, 4)

        return -ax, -ay, -az

    def getPitchAndRoll(self):
        a = self.getAccel()
        pitch = atan2(a[1], -a[2])
        anorm = sqrt(a[0] * a[0] + a[1] * a[1] + a[2] * a[2])
        roll = asin(a[0] / anorm)
        return round(degrees(pitch), 1), round(degrees(roll), 1)
    
    