# hp206c.py
# Version 1.00 - Sep 17, 2018

# Original from Seed Studio (Grove Sensor for Rasperry Pi)
# Slightly adapted by A. Pluess to be used with ESP32, www.aplu.ch

import time
from utils import I2C_com
from machine import Pin

class HP206C:
    HP20X_I2C_DEV_ID       = 0xEC >> 1    #CSB PIN is VDD level(address is 0x76)
    HP20X_I2C_DEV_ID2      = 0XEE >> 1    #CSB PIN is GND level(address is 0x77)
    HP20X_SOFT_RST         = 0x06
    HP20X_WR_CONVERT_CMD   = 0x40
    HP20X_CONVERT_OSR4096  = 0 << 2
    HP20X_CONVERT_OSR2048  = 1 << 2
    HP20X_CONVERT_OSR1024  = 2 << 2
    HP20X_CONVERT_OSR512   = 3 << 2
    HP20X_CONVERT_OSR256   = 4 << 2
    HP20X_CONVERT_OSR128   = 5 << 2

    HP20X_READ_P           = 0x30   #read_p command
    HP20X_READ_A           = 0x31   #read_a command
    HP20X_READ_T           = 0x32   #read_t command
    HP20X_READ_PT          = 0x10   #read_pt command
    HP20X_READ_AT          = 0x11   #read_at command
    HP20X_READ_CAL         = 0X28   #RE-CAL ANALOG

    HP20X_WR_REG_MODE      = 0xC0
    HP20X_RD_REG_MODE      = 0x80

    ERR_WR_DEVID_NACK      = 0x01
    ERR_RD_DEVID_NACK      = 0x02
    ERR_WR_REGADD_NACK     = 0x04
    ERR_WR_REGCMD_NACK     = 0x08
    ERR_WR_DATA_NACK       = 0x10
    ERR_RD_DATA_MISMATCH   = 0x20

    I2C_DID_WR_MASK        = 0xFE
    I2C_DID_RD_MASK        = 0x01

    T_WIN_EN               = 0X01
    PA_WIN_EN              = 0X02
    T_TRAV_EN              = 0X04
    PA_TRAV_EN             = 0X08
    PA_RDY_EN              = 0X20
    T_RDY_EN               = 0X10

    T_WIN_CFG              = 0X01
    PA_WIN_CFG             = 0X02
    PA_MODE_P              = 0X00
    PA_MODE_A              = 0X40

    T_TRAV_CFG             = 0X04

    OK_HP20X_DEV           = 0X80        #HP20x_dev successfully initialized
    REG_PARA               = 0X0F        #Status register

    OSR_CFG = HP20X_CONVERT_OSR1024
    OSR_ConvertTime = 25

    def __init__(self, i2c_address = 0x76, scl = Pin(22), sda = Pin(21)):
        self._i2c = I2C_com(i2c_address, scl, sda)
        self.writeCmd(self.HP20X_SOFT_RST)
        time.sleep(0.1)

    def readTemperature(self):
        self.writeCmd(self.HP20X_WR_CONVERT_CMD|self.OSR_CFG)
        time.sleep(self.OSR_ConvertTime/1000.0)
        t_raw = self._i2c.readBlock(self.HP20X_READ_T, 3)
        t = t_raw[0] << 16 | t_raw[1] << 8 | t_raw[2]
        if t & 0x800000:
            t |= 0xff000000
        return t / 100.0

    def readPressure(self):
        self.writeCmd(self.HP20X_WR_CONVERT_CMD|self.OSR_CFG)
        time.sleep(self.OSR_ConvertTime/1000.0)
        p_raw = self._i2c.readBlock(self.HP20X_READ_P, 3)
        p = p_raw[0] << 16 | p_raw[1] <<8 | p_raw[2]
        if p&0x800000:
            p |= 0xff000000
        return p / 100.0

    def readAltitude(self):
        self.writeCmd(self.HP20X_WR_CONVERT_CMD|self.OSR_CFG)
        time.sleep(self.OSR_ConvertTime/1000.0)
        a_raw = self._i2c.readBlock(self.HP20X_READ_A, 3)
        a = a_raw[0] << 16 | a_raw[1] <<8 | a_raw[2]
        if a&0x800000:
            a|=0xff000000
        return a / 100.0

    def writeCmd(self,uCmd):
        self._i2c.writeTo([uCmd])

