# mcp230xx.py
# MCP23017 and MCP23008 GPIO Expander
# Code from Tony DiCola (Adafruit) and Cefn Hoile (port to ESP8266)
# Pullup 1 k on SDA, SCL needed
# Pullup at reset pin needed
# Address lines must not remain open
# Version 1.02 - Aug 19, 2018

from utils import I2C_com
from machine import Pin

# ----------------------------- class MCP ----------------------------
class MCP():
    PUD_OFF = 0
    PUD_DOWN= 1
    PUD_UP  = 2
    OUT     = 0
    IN      = 1
   
    def __init__(self, scl = Pin(22), sda = Pin(21), addr = 0x20):
        self._i2c_com = I2C_com(addr, scl, sda)
        # Assume starting in ICON.BANK = 0 mode (sequential access)
        # Number of bytes needed to store count of GPIO
        self._nb_gpio_bytes = self.NUM_GPIO // 8
        # Buffer register values so they can be changed without reading
        self._iodir = bytearray(self._nb_gpio_bytes) # Default direction: all inputs
        self._gppu = bytearray(self._nb_gpio_bytes)  # Pullups disabled
        self._gpio = bytearray(self._nb_gpio_bytes)  # All 0
        self._write_iodir(self._iodir)
        self._write_gppu(self._gppu)

    def setup(self, pin, direction, pullup = False):
        """
        Sets the input or output mode for a specified pin. 
        direction should be either MCP.OUT or MCP.IN.
        """
        self._validate_pin(pin)
        # Set bit to 1 for input or 0 for output.
        if direction == MCP.IN:
            self._iodir[pin // 8] |= 1 << (pin % 8)
        elif direction == MCP.OUT:
            self._iodir[pin // 8] &= ~(1 << (pin % 8))
        else:
            raise ValueError('Unexpected value.  Must be IN or OUT.')
        self._write_iodir(self._iodir)
        if direction == MCP.IN:
            self._pullup(pin, pullup)

    def write_digital(self, pin, value):
        """
        Sets the specified pin the provided 0, 1 value. 
        """
        self.write_pins({pin : value})

    def write_pins(self, pins):
        """
        Sets multiple pins high or low at once.  Pins should be a dictonary of pin number and pin value.  
        """
        for pin in pins.keys():
            self._validate_pin(pin)
        for pin, value in iter(pins.items()):
            if value:
                self._gpio[pin // 8] |= 1 << (pin % 8)
            else:
                self._gpio[pin // 8] &= ~(1 << (pin % 8))
        self._write_gpio(self._gpio)

    def read_digital(self, pin):
        """
        Reads the specified pin and return 0, 1.
        """
        return self.read_pins([pin])[0]

    def read_pins(self, pins):
        """
        Reads multiple pins specified in the given list and return list of pin values 0 or 1.
        """
        for pin in pins:
            self._validate_pin(pin)
        # Get GPIO state.
        self._read_gpio()
        # Return True if pin's bit is set.
        return [(self._gpio[pin // 8] & 1 << (pin % 8)) > 0 for pin in pins]

    # ------------ private methods ---------------------------           
    def _pullup(self, pin, enabled):
        if enabled:
            self._gppu[pin // 8] |= 1 << (pin % 8)
        else:
            self._gppu[pin // 8] &= ~(1 << (pin % 8))
        self._write_gppu(self._gppu)
     
    def _write_iodir(self, iodir):
        """
        Write the specified byte value to the IODIR registor.
        """
        self.iodir = iodir
        self._i2c_com.writeBlock(self.IODIR, self.iodir)

    def _write_gppu(self, gppu):
        """
        Writes the specified byte value to the GPPU register.  
        """
        self._gppu = gppu
        self._i2c_com.writeBlock(self.GPPU, self._gppu)

    def _read_gpio(self):
        self._gpio = self._i2c_com.readBlock(self.GPIO, self._nb_gpio_bytes)

    def _write_gpio(self, gpio):
        self._gpio = gpio
        self._i2c_com.writeBlock(self.GPIO, self._gpio)

    def _validate_pin(self, pin):
        if pin < 0 or pin >= self.NUM_GPIO:
            raise ValueError('Invalid GPIO value, must be between 0 and {0}.'.format(self.NUM_GPIO))

# ----------------------------- class MCP23017 ----------------------------
class MCP23017(MCP):
    """
    MCP23017-based GPIO class with 16 GPIO pins.
    """
    NUM_GPIO = 16
    IODIR    = 0x00
    GPIO     = 0x12
    GPPU     = 0x0C


# ----------------------------- class MCP23008 ----------------------------
class MCP23008(MCP):
    """
    MCP23008-based GPIO class with 8 GPIO pins.
    """
    NUM_GPIO = 8
    IODIR    = 0x00
    GPIO     = 0x09
    GPPU     = 0x06

    