# oled.py
# SSD1306 based oled display
# Version 2.00 - Aug 14, 2018

from utils import I2C_com
from machine import Pin
import framebuf, gc

# ------------------------------ class SSD1306 -----------------------------------
class SSD1306:
    _CONTRAST        = const(0x81)
    _ENTIRE_ON       = const(0xa4)
    _NORM_INV        = const(0xa6)
    _DISP            = const(0xae)
    _MEM_ADDR        = const(0x20)
    _COL_ADDR        = const(0x21)
    _PAGE_ADDR       = const(0x22)
    _DISP_START_LINE = const(0x40)
    _SEG_REMAP       = const(0xa0)
    _MUX_RATIO       = const(0xa8)
    _COM_OUT_DIR     = const(0xc0)
    _DISP_OFFSET     = const(0xd3)
    _COM_PIN_CFG     = const(0xda)
    _DISP_CLK_DIV    = const(0xd5)
    _PRECHARGE       = const(0xd9)
    _VCOM_DESEL      = const(0xdb)
    _CHARGE_PUMP     = const(0x8d)

    def __init__(self, width, height, external_vcc):
        self.width = width
        self.height = height
        self.external_vcc = external_vcc
        self.pages = self.height // 8
        self.init_display()

    def init_display(self):
        for cmd in (
            _DISP | 0x00, 
            _MEM_ADDR, 0x00,
            _DISP_START_LINE | 0x00,
            _SEG_REMAP | 0x01,
            _MUX_RATIO, self.height - 1,
            _COM_OUT_DIR | 0x08,
            _DISP_OFFSET, 0x00,
            _COM_PIN_CFG, 0x02 if self.height == 32 else 0x12,
            _DISP_CLK_DIV, 0x80,
            _PRECHARGE, 0x22 if self.external_vcc else 0xf1,
            _VCOM_DESEL, 0x30,
            _CONTRAST, 0xff,
            _ENTIRE_ON,
            _NORM_INV,
            _CHARGE_PUMP, 0x10 if self.external_vcc else 0x14,
            _DISP | 0x01):
                self.write_cmd(cmd)
        self.fill(0)
        self.show()

    def poweroff(self):
        self.write_cmd(_DISP | 0x00)

    def contrast(self, contrast):
        self.write_cmd(_CONTRAST)
        self.write_cmd(contrast)

    def invert(self, invert):
        self.write_cmd(_NORM_INV | (invert & 1))

    def show(self):
        x0 = 0
        x1 = self.width - 1
        if self.width == 64:
            x0 += 32
            x1 += 32
        self.write_cmd(_COL_ADDR)
        self.write_cmd(x0)
        self.write_cmd(x1)
        self.write_cmd(_PAGE_ADDR)
        self.write_cmd(0)
        self.write_cmd(self.pages - 1)
        self.write_framebuf()

    def fill(self, col):
        self.framebuf.fill(col)

    def pixel(self, x, y, col):
        self.framebuf.pixel(x, y, col)

    def scroll(self, dx, dy):
        self.framebuf.scroll(dx, dy)

    def text(self, string, x, y, col=1):
        self.framebuf.text(string, x, y, col)
        
# ------------------------------ class SSD1306_I2C -----------------------------------
class SSD1306_I2C(SSD1306):
    def __init__(self, width, height, scl, sda, addr = 0x3c, external_vcc = False):
        self._i2c_com = I2C_com(addr, scl, sda)
        self.temp = bytearray(2)
        self.buffer = bytearray(((height // 8) * width) + 1)
        self.buffer[0] = 0x40
        self.framebuf = framebuf.FrameBuffer1(memoryview(self.buffer)[1:], width, height)
        super().__init__(width, height, external_vcc)

    def write_cmd(self, cmd):
        self.temp[0] = 0x80
        self.temp[1] = cmd
        self._i2c_com.writeTo(self.temp)

    def write_framebuf(self):
        self._i2c_com.writeTo(self.buffer)

    def dispose(self):
        self._i2c_com.dispose()
        self.buffer = None
        self.framebuf = None
        gc.collect()
#        self.i2c.deinit()    

# ------------------------------ class Oled -----------------------------------
class Oled(SSD1306_I2C):
    def __init__(self, scl = Pin(22), sda = Pin(21), tiny = False):
        SSD1306_I2C.__init__(self, 128, 32 if tiny else 64, scl, sda)
        self.fill(0)

    def line(self, x0, y0, x1, y1):
        # Bresenham algorithm, from Wiki  
        dx =  abs(x1 - x0) 
        sx = 1 if x0 < x1 else -1
        dy = -abs(y1 - y0)
        sy = 1 if y0 < y1 else -1
        err = dx + dy
    
        while True:
            self.pixel(x0, y0, 1)
            if x0 == x1 and y0 == y1: 
                break
            e2 = 2 * err
            if e2 > dy: 
                err += dy
                x0 += sx
            if e2 < dx:
                err += dx
                y0 += sy

    def ellipse(self, xm, ym, a, b):
        # Ellipse algorithm from Wiki  
        dx = 0 
        dy = b 
        a2 = a * a
        b2 = b * b
        err = b2 - (2*b - 1) * a2

        while True:
            self.pixel(xm + dx, ym + dy, 1)
            self.pixel(xm - dx, ym + dy, 1)
            self.pixel(xm - dx, ym - dy, 1)
            self.pixel(xm + dx, ym - dy, 1)

            e2 = 2*err;
            if (e2 <  (2 * dx + 1) * b2):
                dx += 1 
                err += (2 * dx + 1) * b2
            if (e2 > -(2 * dy - 1) * a2):
                dy -= 1 
                err -= (2 * dy - 1) * a2
            if dy < 0:
               break

    def circle(self, xm, ym, radius):
        self.ellipse(xm, ym, radius, radius)
   
    def rectangle(self, xm , ym, a, b):
        self.framebuf.rect(xm - a // 2, ym - b // 2, a, b, 1)

    def fillRectangle(self, xm, ym, a, b):
        self.framebuf.fill_rect(xm - a // 2, ym - b // 2, a, b, 1)

    def image(self, filename):
        # 3 header lines assumed: 
        #    PPM identifier (e.g, P3 or P6)<cr>
        #    width height<cr>
        #    format e.g. 255<cr>
        #    comment lines starting with # are not considered
        inHeader = True
        n = 0
        try:
            with open(filename, 'rb') as f:
                while inHeader:
                    line = f.readline()
                    if line[0] == '#': # comment
                        continue
                    n += 1
                    if n == 4:
                        inHeader = False
                # data now
                x = 0
                y = 0
                while True:
                    b = f.read(3) # get rgb color triple as bytes
                    if len(b) == 0:
                        break # eof
                    if b[0] != 0: # only test 1 component
                        self.framebuf.pixel(x, y, 1)
                    x += 1    
                    if x == 128:
                        x = 0
                        y += 1
        except:
            print("Failed to load picture '" + filename + "'")   
