# oxoaccelerometer.py
# Version 1.15 - Aug 17, 2018
# LIS3DE accelerometer

from time import sleep
from utils import I2C_com
from machine import Pin
import _thread
from math import atan2, asin, sqrt, degrees

class Accelerometer:
    g = 9.81 
    I2C_ADDRESS     = const(0x28)
    ACCEL_INT_PIN    = const(35)  # Interrupt pin
    POLLING_INTERVAL  = 0.05

    STATUS_REG_AUX  = const(0x07)
    OUT_ADC1_L      = const(0x08)
    OUT_ADC1_H      = const(0x09)
    OUT_ADC2_L      = const(0x0A)
    OUT_ADC2_H      = const(0x0B)
    OUT_ADC3_L      = const(0x0C)
    OUT_ADC3_H      = const(0x0D)
    INT_COUNTER_REG = const(0x0E)
    WHO_AM_I        = const(0x0F)
    TEMP_CFG_REG    = const(0x1F)
    CTRL_REG1       = const(0x20)
    CTRL_REG2       = const(0x21)
    CTRL_REG3       = const(0x22)
    CTRL_REG4       = const(0x23)
    CTRL_REG5       = const(0x24)
    CTRL_REG6       = const(0x25)
    REFERENCE       = const(0x26)
    STATUS_REG2     = const(0x27)
    ACCELERATIONS   = const(0x28)  # reserved
    OUT_X           = const(0x29)
    OUT_Y           = const(0x2B)
    OUT_Z           = const(0x2D)
    FIFO_CTRL_REG   = const(0x2E)
    FIFO_SRC_REG    = const(0x2F)
    IG1_CFG         = const(0x30) # Interrupt generator 1
    IG1_SOURCE      = const(0x31)
    IG1_THS         = const(0x32)
    IG1_DURATION    = const(0x33)
    IG2_CFG         = const(0x34) # Interrupt generator 2
    IG2_SOURCE      = const(0x35)
    IG2_THS         = const(0x36)
    IG2_DURATION    = const(0x37)
    CLICK_CFG       = const(0x38)
    CLICK_SRC       = const(0x39)
    CLICK_THS       = const(0x3A)
    TIME_LIMIT      = const(0x3B)
    TIME_LATENCY    = const(0x3C)
    TIME_WINDOW     = const(0x3D)
    Act_THS         = const(0x3E)
    Act_DUR         = const(0x3F)
    
    MODE_NORMAL       = const(0)
    DATA_RATE_0HZ     = const(0x00) # Power-down mode
    DATA_RATE_1HZ     = const(0x01)
    DATA_RATE_10HZ    = const(0x02)
    DATA_RATE_25HZ    = const(0x03)
    DATA_RATE_50HZ    = const(0x04)
    DATA_RATE_100HZ   = const(0x05)
    DATA_RATE_200HZ   = const(0x06)
    DATA_RATE_400HZ   = const(0x07)
    DATA_RATE_1620HZ  = const(0x08)  # works only in LOW_POWER mode
    DATA_RATE_1344HZ  = const(0x09)  # works only in NORMAL mode
    DATA_RATE_5376HZ  = const(0x09)  # works only in LOW_POWER mode
    
    RANGE_2G  = const(0b00)   # +/- 2g (default value)
    RANGE_4G  = const(0b01)   # +/- 4g
    RANGE_6G  = const(0b10)   # +/- 8g
    RANGE_16G = const(0b11)   # +/- 16g
    
    SINGLE_CLICK_X = const(0x01)
    DOUBLE_CLICK_X = const(0x02)
    SINGLE_CLICK_Y = const(0x04)
    DOUBLE_CLICK_Y = const(0x08)
    SINGLE_CLICK_Z = const(0x10)
    DOUBLE_CLICK_Z = const(0x20)
    
    _instance = None
    _click_callback = None
    _trigger_level = 0
    _rearm_time = 0
    _click_config = 0

    def __init__(self, *args):
        if len(args) != 1 or args[0] != "$$$xxxprivate":
            raise Exception("Use factory method create() to return a single instance")
        if Accelerometer._instance == None:
            self._irPin = Pin(Accelerometer.ACCEL_INT_PIN, Pin.IN)
            self._isRunning = True
            _thread.start_new_thread(self._run, ())
            sleep(0.1) # necessary startup time after power-up
            self._i2c_com = I2C_com(Accelerometer.I2C_ADDRESS)
            self._setDataRate(Accelerometer.DATA_RATE_400HZ)
            self._setMode(Accelerometer.MODE_NORMAL)
            self._setRange(Accelerometer.RANGE_2G)
            self._enableClick(min(127, int(Accelerometer._trigger_level * 1.27)))
            self._enableClickInterrupt(True)
            self._enableTemp()
        self._resetClick()

    # ----------- public methods ---------------------
    @staticmethod
    def create(click_callback = None, **kwargs):
        Accelerometer._click_callback = click_callback
        Accelerometer._trigger_level = 10
        Accelerometer._rearm_time = 0.2
        Accelerometer._click_config = Accelerometer.SINGLE_CLICK_X | Accelerometer.SINGLE_CLICK_Y | Accelerometer.SINGLE_CLICK_Z
        for key in kwargs:
            if key == 'trigger_level': # in range 1..100,  1: highest sensitivity, 100: lowest sensitiviy; default: 10
                Accelerometer._trigger_level = max(1, kwargs[key])
            if key == 'rearm_time': # time in s until sensor is rearmed; default: 0.2
                Accelerometer._rearm_time = kwargs[key]
            if key == 'click_config': # click configuration; default CLICK_X | CLICK_Y | CLICK_Z
                Accelerometer._click_config = kwargs[key]
        if Accelerometer._instance == None:
            Accelerometer._instance = Accelerometer("$$$xxxprivate")
        return Accelerometer._instance

    def getRaw(self):
        return self._i2c_com.readBlock(Accelerometer.ACCELERATIONS | 0x80, 6) # sets autoincrement bit
    
    def getValues(self):
        v = self.getRaw()
        x = self._convert(v[0], v[1]) / 17800 * Accelerometer.g
        y = self._convert(v[2], v[3]) / 17800 * Accelerometer.g
        z = self._convert(v[4], v[5]) / 17800 * Accelerometer.g
        return round(x, 2), round(y, 2), round(z, 2)

    def getX(self):
        v = self.getRaw()
        x = self._convert(v[0], v[1]) / 17800 * Accelerometer.g
        return round(x, 2)

    def getY(self):
        v = self.getRaw()
        y = self._convert(v[2], v[3]) / 17800 * Accelerometer.g
        return round(y, 2)
    
    def getZ(self):
        v = self.getRaw()
        z = self._convert(v[4], v[5]) / 17800 * Accelerometer.g
        return round(z, 2)

    def getPitch(self):
        a = self.getValues()
        pitch = atan2(a[1], a[2])
        return -int(degrees(pitch))
    
    def getRoll(self):
        a = self.getValues()
        anorm = sqrt(a[0] * a[0] + a[1] * a[1] + a[2] * a[2])
        if anorm == 0:
            return 0
        roll = asin(a[0] / anorm)
        return int(degrees(roll))

    def wasClicked(self):
        sleep(0.01)  # in case of narrow loops
        tmp = self._clicked
        self._clicked = False
        return tmp
    
    def getTemperature(self):
        temp_lo = self._i2c_com.readReg(Accelerometer.OUT_ADC3_L)
        t = self._i2c_com.readReg(Accelerometer.OUT_ADC3_H)
        if t > 127:
            t = t - 256  # 2's complement
        return t + 20

    def dispose(self):
        if Accelerometer._instance != None:
            Accelerometer._instance._i2c_com.dispose()
        self._isRunning = False

    # -------------- private methods ---------------------
    def _run(self):
        # Sensor thread
        while self._isRunning:
            try:
                if self._irPin.value() == 1:
                    self._clicked = True
                    if Accelerometer._click_callback != None:
                        Accelerometer._click_callback()
                    sleep(Accelerometer._rearm_time)    
                    self._checkClick()  # reset pending interrupt request
                sleep(Accelerometer.POLLING_INTERVAL)
            except Exception as ex:
                self._isRunning = False
                print("Exception in sensor polling thread: " + str(ex))        
                
    '''                
    def _getChipID(self):
        return self._i2c_com.readReg(Accelerometer.WHO_AM_I)
    '''
    
    def _setMode(self, mode):
        reg1 = self._i2c_com.readReg(Accelerometer.CTRL_REG1)
        reg1 &= ~0x08
        reg1 |= (mode << 3)
        self._i2c_com.writeReg(Accelerometer.CTRL_REG1, reg1)
    
    def _enableTemp(self):
        temp_enable = 0xC0
        self._i2c_com.writeReg(Accelerometer.TEMP_CFG_REG, temp_enable)
        reg4 = self._i2c_com.readReg(Accelerometer.CTRL_REG4)
        reg4 |= 0x80
        self._i2c_com.writeReg(Accelerometer.CTRL_REG4, reg4)
     
    '''       
    def _getMode(self):
        return (self._i2c_com.readReg(Accelerometer.CTRL_REG1) & 0x08) >> 3
    '''
    
    def _setDataRate(self, dr):
        reg1 = self._i2c_com.readReg(Accelerometer.CTRL_REG1)
        reg1 &= ~0xF0
        reg1 |= (dr << 4)
        self._i2c_com.writeReg(Accelerometer.CTRL_REG1, reg1)
    
    '''
    def _getDataRate(self):
        return (self._i2c_com.readReg(Accelerometer.CTRL_REG1) & 0xF0) >> 4
    '''
    
    def _setRange(self, range):
        reg4 = self._i2c_com.readReg(Accelerometer.CTRL_REG4)
        reg4 &= ~0x30
        reg4 |= (range << 4)
        self._i2c_com.writeReg(Accelerometer.CTRL_REG4, reg4)
    
    '''
    def _getRange(self):
        return (self._i2c_com.readReg(Accelerometer.CTRL_REG4) & 0x30) >> 4
    '''
   
    def _convert(self, lsb, msb):
        value = lsb | (msb << 8)
        if value & 0x8000:
          value = value - 65536
        return value        

    def _enableClick(self, threshold):
#        print("click_callback:", Accelerometer._click_callback, "trigger_level:", Accelerometer._trigger_level,
#          "rearm_time:", Accelerometer._rearm_time,  "click_config:", hex(Accelerometer._click_config))

        self._i2c_com.writeReg(Accelerometer.CLICK_CFG, Accelerometer._click_config)
 
        ths = 0x80 | threshold # LIR (Latch Interrupt Request) bit (MSB) set
        self._i2c_com.writeReg(Accelerometer.CLICK_THS, ths)
        
        click_time_limit = 50  # unit?  0..127 higher values inhibit click events for a longer time
        self._i2c_com.writeReg(Accelerometer.TIME_LIMIT, click_time_limit)

        double_click_time_latency = 10 # unit?
        self._i2c_com.writeReg(Accelerometer.TIME_LATENCY, double_click_time_latency)
        
        double_click_time_window = 255 # unit?
        self._i2c_com.writeReg(Accelerometer.TIME_WINDOW, double_click_time_window)

    def _resetClick(self):    
        self._clicked = False
        self._checkClick()  # reset pending interrupt request

    def _checkClick(self):
        # also resets pending interrupt request
        return self._i2c_com.readReg(Accelerometer.CLICK_SRC)

    def _enableClickInterrupt(self, enable):
        if enable:
            self._i2c_com.writeReg(Accelerometer.CTRL_REG3, 0x80)
        else:
            self._i2c_com.writeReg(Accelerometer.CTRL_REG3, 0)

