# pcf857x.py
# PCF8574 and PCF8575  GPIO Expander
# Version 2.00 - Aug 14, 2018

from utils import I2C_com
from machine import Pin

# ------------------------------ class PCF8574 -----------------------------------
class PCF8574:
    def __init__(self, scl = Pin(22), sda = Pin(21), addr = 0x38):
        self._i2c_com = I2C_com(addr, scl, sda)
        self._input = 0  # Buffers the result of read in memory
        self._input_mask = 0  # Mask specifying which pins are set as input
        self._output = 0  # The state of pins set for output
        self.set_all() # all inputs

    def _read(self):
        self._input = self._i2c_com.readFrom(1)[0] & self._input_mask

    def _write(self):
        self._i2c_com.writeTo(bytes([self._output | self._input_mask]))

    def read_digital(self, pin):
        bit_mask = 1 << pin
        self._input_mask |= bit_mask
        self._output &= ~bit_mask
        self._write()  # Update input mask before reading
        self._read()
        return (self._input & bit_mask) >> pin

    def read_all(self):
        self._input_mask = 0xFF
        self._output = 0
        self._write()  # Update input mask before reading
        self._read()
        return self._input
    
    def read_all_inv(self):
        v = self.read_all()
        return ~v & 0xFF

    def write_digital(self, pin, value):
        bit_mask = 1 << pin
        self._input_mask &= ~bit_mask
        self._output = self._output | bit_mask if value else self._output & (~bit_mask)
        self._write()

    def write_all(self, value):
        self._input_mask = 0
        self._output = value
        self._write()

    def write_all_inv(self, value):
        self.write_all(~value)    

    def set_all(self):
        self.write_all(0xFF)

    def clear_all(self):
        self.write_all(0x0)

    def toggle(self, pin):
        bit_mask = 1 << pin
        self._input_mask &= ~bit_mask
        self._output ^= bit_mask

# ------------------------------ class PCF8575 -----------------------------------
class PCF8575:
    def __init__(self, scl = Pin(22), sda = Pin(21), addr = 0x20):
        self._i2c_com = I2C_com(addr, scl, sda)
        self._input = 0  # Buffers the result of read in memory
        self._input_mask = 0  # Mask specifying which pins are set as input
        self._output = 0  # The state of pins set for output
        self.set_all() # all inputs

    def _read(self):
        data = self._i2c_com.readFrom(2) # byte order: low-high (little endian)
        v = (data[1] << 8) + data[0]
        self._input = v & self._input_mask

    def _write(self):
        out = self._output | self._input_mask
        low = out & 0xff
        high = (out >> 8) & 0xff
        self._i2c_com.writeTo([low, high])  # byte order: low-high (little endian)

    def read_digital(self, pin):
        bit_mask = 1 << pin
        self._input_mask |= bit_mask
        self._output &= ~bit_mask
        self._write()  # Update input mask before reading
        self._read()
        return (self._input & bit_mask) >> pin

    def read_all(self):
        self._input_mask = 0xFFFF
        self._output = 0
        self._write()  # Update input mask before reading
        self._read()
        return self._input

    def read_all_inv(self):
        v = self.read_all()
        return ~v & 0xFFFF

    def write_digital(self, pin, value):
        bit_mask = 1 << pin
        self._input_mask &= ~bit_mask
        self._output = self._output | bit_mask if value else self._output & (~bit_mask)
        self._write()

    def write_all(self, value):
        self._input_mask = 0
        self._output = value
        self._write()
        
    def write_all_inv(self, value):
        self.write_all(~value)    

    def set_all(self):
        self.write_all(0xFFFF)

    def clear_all(self):
        self.write_all(0x0)

    def toggle(self, pin):
        bit_mask = 1 << pin
        self._input_mask &= ~bit_mask
        self._output ^= bit_mask

