# radio.py
# Version 1.07, Feb. 9, 2019
 
import gc
from tcpcom import *
from time import sleep
import _thread
 
class Radio:
    _bufsize = 100
    _serverIP = "192.168.4.1"
    _port = 1055
 
    def __init__(self, id, verbose = False):
        if not (id == "A" or id == "B"):
            print("Illegal value for node")
            return
        self._id = id 
        self._isConnected = False
        self._verbose = verbose
        self._isRunning = False
 
    def on(self):
        self._debug("Turning radio on...")
        self._buffer = RingBuffer(Radio._bufsize)
        if self._id == 'B': # start server
            self._listen() 
        elif self._id == 'A':  # try to connect with client
            _thread.start_new_thread(self._run, ())
     
    def _listen(self, ssid = "$oxocard"):
        self._debug("Activating AP")
        Wlan.activateAP(ssid = ssid, password = "")
        self._debug("Starting TCPServer at port " + str(Radio._port))
        self._server = TCPServer(Radio._port, stateChanged = self._onStateChanged, endOfBlock = '\n', 
                                 verbose = self._verbose)
 
    def _connect(self, ssid = "$oxocard"):
        if not Wlan.connect(ssid, ""):
            self._debug("Connection to AP failed")
            return False
        self._client = TCPClient(verbose = self._verbose)
        self._debug("Trying to connect to host " + Radio._serverIP + " at port " + str(Radio._port))
        if not self._client.connect(Radio._serverIP, Radio._port):
            self._debug("Connection failed.")
            return False
        self._isConnected = True
        return True
     
    def send(self, msg):
        sleep(0.01) # avoid narrow loops
        if not self._isConnected:
            return
        try:
            self._debug("Sending msg: " + str(msg))
            if self._id == 'B':
                self._server.send(str(msg) + "\n")
            if self._id == 'A':
                self._client.send(str(msg) + "\n")
        except:
            return False
        return True
         
    def receive(self):
        sleep(0.01) # avoid narrow loops
        if not self._isConnected:
            return
        item = self._buffer.get()
        if item == None:
            return ""
        self._debug("Receiving msg: " + str(item))
        return item
 
    def off(self):
        self._debug("Releasing resources")
        self._isRunning = False
        self._isConnected = False
        if self._id == 'B':
            self._server.terminate()
        if self._id == 'A':
            self._client.closeConnection()
        self._buffer = None
        gc.collect()
         
    def isConnected(self):
        sleep(0.01) # avoid narrow loops
        return self._isConnected
     
    def waitConnection(self):
        while not self.isConnected():
            pass
         
    def _onStateChanged(self, state, msg):                            
        if state == "LISTENING":
            self._debug("Waiting for client connection")
        elif state == "PORT_IN_USE":
            self._debug("PORT_IN_USE")
        elif state == "TERMINATED":
            self._debug("TERMINATED")
        elif state == "CONNECTED":
            self._debug("Client " + msg + " connected")
            self._isConnected = True
        elif state == "MESSAGE":
            self._debug("Got msg: " + msg)
            if self._buffer != None:
                self._buffer.put(msg)
     
    def _debug(self, msg):
        if self._verbose:
            print("    Debug:-- " + msg)
             
    def _run(self):
        success = False
        while not success:
            success = self._connect()       
            sleep(2)
        self._isRunning = True
        while self._isRunning:
            try:    
                msg = self._client.getReply()
                self._debug("Client got reply: " + str(msg))
                msg = msg.decode('utf-8')
                self._buffer.put(str(msg)[:-1]) # remove \n
            except:
                self._debug("Exception in receive thread")
        self._debug("Thread terminated") 
 
# -------------- class RingBuffer -----------------
class RingBuffer:
    def __init__(self, size = 100):
        self.buffer = [None] * size
        self.size = size
        self.put_ptr = 0
        self.get_ptr = 0
        self.count = 0
 
    def put(self, data):
        if self.count < self.size:
            p = self.put_ptr
            self.buffer[p] = data
            self.put_ptr = (p + 1) % self.size
            self.count += 1
 
    def get(self):
        if self.count > 0:
            p = self.get_ptr
            data = self.buffer[p]
            self.get_ptr = (p + 1) % self.size
            self.count -= 1
            return data
        return None
 
    def __len__(self):
        return self.count
 
    def __getitem__(self, index):
        if 0 <= index < self.count:
            p = (self.get_ptr + index) % self.size
            return self.buffer[p]
        return None