# tcpcom.py
# Version 1.12 - Feb 16, 2019

import _thread
import network
import socket
import ussl
from time import sleep
import sys
import ujson

# ================================== class Wlan ============================
class Wlan:
    '''
    All methods are static.
    '''
    ap_connected = False
    wlanIF = None
    isAP_activated = False
    wlanAP = None
    verbose = False
    
    # -------------------- connect to external access point -------------------
    @staticmethod
    def connect(ssid, pwd, timeout = 10, verbose = False):
        '''
        If the server is not yet connected, performs a connection trial to 
        the access point with given ssid and password. The connection trial
        is performed during the given timeout (default: 10 s).
        Returns True, if connection is established; otherwise False
        '''
        # Wait a moment to be sure that other hardware is up and running (e.g. neopixels)
        sleep(1)
        Wlan.verbose = verbose
        # Deactivate AP if it is active
        try:
            ap = network.WLAN(network.AP_IF)
            if ap.active():
                Wlan._debug("Must deactivate access point")
                ap.active(False)
        except: # WiFi already started in different mode
            pass     
        if Wlan.ap_connected: # do not connect twice
            return True    
        Wlan.wlanIF = network.WLAN(network.STA_IF)
        Wlan.wlanIF.active(True)
        Wlan._debug("Check if already connected")
        if not Wlan.wlanIF.isconnected():
            Wlan._debug("Calling wlan.connect()")
            Wlan.wlanIF.connect(ssid, pwd)
            trials = 0
            while (not Wlan.wlanIF.isconnected()) and trials < timeout:
                Wlan._debug("Connection trial #" + str(trials))
                sleep(1)
                trials += 1
            if trials == 10:
                Wlan._debug("Timeout while connecting")
                Wlan.wlanIF.active(False)
                Wlan.ap_connected = False
            else:
                Wlan._debug("Connection to AP successful")
                Wlan.ap_connected = True
        else:
            Wlan._debug("Connection already established")
            Wlan.ap_connected = True
        return Wlan.ap_connected
    
    @staticmethod
    def isWlanConnected():
        '''
        Returns True, if connection to access point is established; otherwise False.
        '''
        return Wlan.ap_connected
    
    @staticmethod
    def disconnect():
        '''
        Disconnect from access point.
        '''
        if Wlan.wlanIF != None:
            Wlan.wlanIF.active(False)
            Wlan.ap_connected = False

    @staticmethod
    def getIfConfig():
        '''
        Returns the tuple (IP address, mask, gateway IP); None, if not connected.
        '''
        if Wlan.ap_connected:
            return Wlan.wlanIF.ifconfig()
        return None

    @staticmethod
    def getMyIPAddress():
        '''
        Returns the IP address received from router's DHCP server or 
        empty string, if not connected.
        '''
        if Wlan.ap_connected:
            return Wlan.wlanIF.ifconfig()[0]
        return ""

    @staticmethod
    def scan():
        '''
        Returns a tuple of network information about all visible access points.
        Format of each entry: 
        Tuple with ssid (string), channel (int), 
            received signal strength indicatior (rssi) in dB (int).
        '''
        Wlan._debug("Scanning SSIDs")
        ap = network.WLAN(network.STA_IF)
        activated = ap.active()
        if not activated:
            ap.active(True)
        ssids = ap.scan()
        if not activated:
            ap.active(False)
        Wlan._debug("scanWLAN returns " + str(ssids))
        data = []
        for ssid in ssids:
            data.append((ssid[0].decode('utf-8'), ssid[2], ssid[3]))
        return tuple(data)
    
    # -------------------- use internal access point -------------------
    @staticmethod
    def activateAP(ssid = "ESP", password = "aabbaabbaabb", channel = 10, verbose = False):
        Wlan.verbose = verbose
        wlanAP = network.WLAN(network.AP_IF)
        wlanAP.active(True) # do it before config
        if password == "":
            wlanAP.config(essid = ssid, channel = channel, authmode = 1)
            Wlan._debug("Activating AP with SSID: '" + wlanAP.config('essid') + "'; no password; channel: " + str(wlanAP.config('channel')))
        else:
            wlanAP.config(essid = ssid, password = password, channel = channel, authmode = 4)
            Wlan._debug("Activating AP with SSID: '" + wlanAP.config('essid') + "'; password: '" + wlanAP.config('password') + "'; channel: " + str(wlanAP.config('channel')))
        isAP_activated = True
         
    @staticmethod
    def deactivateAP():
        wlanAP.active(False)

    @staticmethod
    def getMacAddress():
        if wlanAP != None:
            return wlanAP.config('mac')
        return ""
    
    @staticmethod
    def setVerbose(enable):
        Wlan.verbose = enable

    @staticmethod
    def _debug(msg):
        if Wlan.verbose:
            print("    Debug:-- " + msg)

# ================================== class TCPServer ================================
class TCPServer():
    ''' 
    Abstraction of a blocking TCP server using a thread.
    '''
    
    PORT_IN_USE = "PORT_IN_USE"
    CONNECTED = "CONNECTED"
    LISTENING = "LISTENING"
    TERMINATED = "TERMINATED"
    MESSAGE = "MESSAGE"

    def __init__(self, port, stateChanged, endOfBlock = "\n", verbose = False):
        ''' 
        Creates a server instance that will listen on given IP port. 
        '''
        self._port = port
        self._endOfBlock = endOfBlock
        self._stateChanged = stateChanged
        self._verbose = verbose
        self._conn = None
        self._bind()
        self._isServerRunning = True
        _thread.start_new_thread(self._run, ())
       
    def _bind(self):
        ''' 
        Binds the server to the port. Returns True, if successful.
        ''' 
        host = ''   # use all available interfaces
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)  # close port when process exits
        self._debug('Socket created. Bind to port ' + str(self._port) + ' now.')
        try:
            self.sock.bind((host, self._port))
        except: 
            self._debug('Bind failed.')
            self._stateChanged(TCPServer.PORT_IN_USE, str(self._port))
            return False
        self.sock.listen(1)  # queued requests are set to 1 (or max for system)
        self._debug('Socket bind successful.')
        return True
    
    def _run(self):
        while self._isServerRunning:
            self._stateChanged(TCPServer.LISTENING, "")
            self._clientIP = ""
            self._accept()  # blocking
            if self._clientIP != "":
                self._debug("Client accepted. IP: " + self._clientIP)
                self._stateChanged(TCPServer.CONNECTED, self._clientIP)
                self._receive()  # blocking
        self._stateChanged(TCPServer.TERMINATED, "")
        self._isServerRunning = False
  
    def _accept(self):
        '''
        Blocking method that listens for an incoming TCP client on selected IP port. Returns the IP address of the caller or empty
        string, if an error occured.
        '''
        self._debug("Calling accept()")
        self._clientIP = ""
        try:
            self._conn, addr = self.sock.accept()
        except Exception as ex:
            self._debug("Exception in accept(): " + str(ex))
            self._isServerRunning = False
            return
        self._clientIP = addr[0]
    
    def closeConnection(self):
        '''
        Closes the TCP connection.
        '''
        self._debug("Calling closeConnection()")
        if self._conn != None:
            self._conn.close()

    def closeSocket(self):
        '''
        Closes the socket.
        '''
        self._debug("Calling closeSocket()")
        self._isServerRunning = False
        self.sock.close()
        
    def terminate(self):
        '''
        Closes the connection and the socket.
        '''
        self.closeConnection()
        self.closeSocket()    
        
    def send(self, data):
        '''
        Sends all given data (normally a bytes type, but string type is also supported). Blocks until
        all data is sent. If the sending fails (e.g. because the client disconnected), returns immediately.
        '''
#        self._debug("Calling send() with data: " + data)
        try:
            self._conn.sendall(data)
        except Exception as ex:
            self._debug("Exception in sendall(): " + str(ex))
            
    def _receive(self):
        '''
        Returns all available data in the receiver buffer as bytes type; Empty, if fails
        '''    
        self._debug("Calling _receive()")
        try:
            while True:
                data = ""
                isReceiving = True
                while not data[-len(self._endOfBlock):] == self._endOfBlock:
                    self._debug("Calling blocking conn.recv()")
                    reply = self._conn.recv(512)  # blocking, buffer rather small to avoid memory overflow
                    if reply == None or len(reply) == 0: # Client disconnected
                        self._debug("conn.recv() returned None")
                        isReceiving = False
                        break
                    data += reply.decode('utf-8')
                    self._debug("Received msg: " + data + " - len: " + str(len(data)))
                if not isReceiving:
                    break
                self._stateChanged(TCPServer.MESSAGE, data[0:-len(self._endOfBlock)]) # do not include endOfBlock
        except Exception as ex:  # May happen if client peer is resetted
            self._debug("_receive() terminated with exception from blocking conn.recv(): " + str(ex))
            return
        self._debug("_receive() terminated because client disconnected")
        
    def isServerRunning(self):
        '''
        Returns True, if the server thread is running.
        '''
        return self._isServerRunning()

    def getClientIP(self):
        '''
        Returns the IP address string of a connected client; empty if no client is connected.
        '''
        return self._clientIP
    
    def _debug(self, msg):
        if self._verbose:
            print("    Debug:-- " + msg)


# ================================== class TCPClient ================================
class TCPClient:
    def __init__(self, verbose = False):
        '''
        Creates a client instance, but does not yet connect to any server.
        '''
        self.connected = False
        self.verbose = verbose
        self._isSSL = False

    def connect(self, host, port):
        '''
        Performs a connection trial to the host (host name or dotted ip address as string) at given port (as integer).
        The method is blocking until the connection is established and returns True. If the connection fails after a timeout of about 20s,
        it returns False.
        '''
        self._connect(host, port)
        return self.connected    

    def _connect(self, host, port):
        self.host = host
        self.port = port
        if self.connected:
            return True
        try:
            addr = socket.getaddrinfo(host, port)[0][-1]
            self._debug("Connecting to " + host + ", port: " + str(port))
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM) 
            self.sock.connect(addr)
            if port == 443:
                self._isSSL = True
                self.sock = ussl.wrap_socket(self.sock)
        except Exception as e:
            self._debug("Connection failed." + str(e))
            return
        self.sock.setblocking(0)
        self.connected = True

    def send(self, msg):
        '''
        Sends the string message. Blocks until all characters are sent.
        '''
        if not self.connected:
            raise Exception('TCPClient failed to send data')
        try:
            if self._isSSL:
                self.sock.write(msg)
            else:
                self.sock.sendall(msg)
        except:
            raise Exception('TCPClient failed to send data')

    def getReply(self, timeout = 5):
        '''
        Waits for the reply and returns it. If no reply is received within the given timeout (default: 5s), returns None.
        '''
        if not self.connected:
            return b''
        nbRetries = 0
        while True:
            try:
                if self._isSSL:
                    reply = self.sock.read(512)
                    if reply == None:
                        raise Exception()
                else:
                    reply = self.sock.recv(512)
                # len = 0, if server closed connection or died
                return reply
            except:  # no reply
                if timeout > 0:
                    nbRetries += 1
                    if nbRetries == 10 * timeout:
                        return None  # Timeout
                sleep(0.1)

    def closeConnection(self):
        '''
        Closes the TCP connection.
        '''
        self._debug("Closing connection")
        if self.connected:
            self.sock.close()
            self.connected = False

    def isConnected(self):
        '''
        Returns True, if connected.
        '''
        return self.connected

  
    def _debug(self, msg):
        if self.verbose:
            print("    Debug:-- " + msg)


# ================================== class HTTPServer ================================
class HTTPServer(TCPServer):  
 
    def getHeader1(self):
       return "HTTP/1.1 501 OK\r\nServer: " + self._serverName + "\r\nConnection: Closed\r\n\r\n"
   
    def getHeader2(self):
        return "HTTP/1.1 200 OK\r\nServer: " + self._serverName + "\r\nContent-Length: %d\r\nContent-Type: text/html\r\nConnection: Closed\r\n\r\n"
     
    def __init__(self, requestHandler, serverName = "ESP", port = 80, verbose = False):
        '''
        Creates a special purpose TCPServer that listens infinitely for a connecting client on given port (default = 80) and parses the incomming message for a 
        HTTP GET request. 
        '''
        TCPServer.__init__(self, port, self._stateChanged, endOfBlock = "\r\n\r\n", verbose = verbose)
        self._serverName = serverName
        self._requestHandler = requestHandler
    
    def getClientIP(self):
        '''
        Returns the dotted IP of a connected client. If no client is connected, returns empty string.
        '''
        return self._clientIP

    def _stateChanged(self, state, msg):
        self._debug("_stateChanged() with state: " + state + " - msg: " + msg)
        if state == TCPServer.MESSAGE:
            filename, params = self._parseURL(msg)
            if filename == None:
                self.send(self.getHeader1())
            else:
                try:
                    self._debug("Invoking requestHandler() with\n       client IP: %s\n       filename: %s\n       params: %s" %(self._clientIP, filename, str(params))) 
                    text = self._requestHandler(self._clientIP, filename, params)
                    if text == None or len(text) == 0:
                        self.send(self.getHeader1())
                    else:    
                        self.send(self.getHeader2() % (len(text)))
                    self.send(text)
                except Exception as e:
                    print(str(e))
            self.closeConnection()

    def _parseURL(self, request):
        try:
            req = str(request, 'utf-8') # bytes to str
        except:  # Can't decode utf-8 
            return None, tuple([])   
        lines = req.split('\n') # split lines
        params = []
        for line in lines:
            if line[0:4] == 'GET ': # only GET request
                url = line.split()[1].strip() # split at spaces and take second item
                i = url.find('?')  # check for params
                if i != -1: # params given
                    filename = url[0:i].strip()  # includes leading /
                    params = []
                    urlParam = url[i + 1:]
                    for item in urlParam.split('&'): # split parameters
                        i = item.find('=')
                        key = item[0:i]
                        value = item[i+1:]
                        params.append((key, value))
                    return filename, tuple(params)    
                return url.strip(), tuple([])
        return None, tuple([])   
    

# ================================== class HTTPClient ================================
class HTTPClient(TCPClient):
    def __init__self(self, verbose = False):
        '''
        Creates a client instance, but does not yet connect to any server.
        '''
        TCPClient.__init_(self, verbose)
        self.verbose = verbose

    def sendGetRequest(self, query, timeout = 10):
        '''
        Sends a HTTP GET request with given query in format <filename>?key1=value1&key2=value2&...
        Waits for timeout seconds for a reply (default: 10).
        '''
        if not self.connected:
            return ""
        request = "GET " + query + " HTTP/1.1\r\nHost: " + self.host + "\r\n\r\n" 
        self._debug("Sending HTTP GET request: " + request.replace('\n', '<lf>').replace('\r', '<cr>'))
        self.send(request)    
        response = ""
        first = True
        while True:
            if first:
                reply = self.getReply(timeout) 
                first = False
            else:
                reply = self.getReply(2) # decrease timeout after first reply
            if reply == None: # Timeout
                self._debug("Timeout in getReply()")
                break
            elif len(reply) == 0:  # Server closed connection
                self._debug("Server closed connection")
                break
            response += str(reply, 'utf-8')
        self._debug("Server response:\n" + response)
        self.closeConnection() # Release resources
        return response
    
    def sendPostRequest(self, query, content, timeout = 10):
        '''
        Sends a HTTP POST request with given data string.
        Waits for timeout seconds for a reply (default: 10).
        '''
        if not self.connected:
            return ""
        request = "POST " + query + " HTTP/1.1\r\nHost: " + self.host + "\r\n" 
        request += 'Content-Type: application/x-www-form-urlencoded\r\n'
        request += 'Content-Length: %s\r\n\r\n%s\r\n\r\n' % (len(content), content)
        self._debug("Sending HTTP POST request: " + request.replace('\n', '<lf>').replace('\r', '<cr>'))
        self.send(request)    
        response = ""
        first = True
        while True:
            if first:
                reply = self.getReply(timeout) 
                first = False
            else:
                reply = self.getReply(2) # decrease timeout after first reply
            if reply == None: # Timeout
                self._debug("Timeout in getReply()")
                break
            elif len(reply) == 0:  # Server closed connection
                self._debug("Server closed connection")
                break
            response += str(reply, 'utf-8')
        self._debug("Server response:\n" + response)
        self.closeConnection() # Release resources
        return response

    def sendDeleteRequest(self, resource, timeout = 10):
        '''
        Sends a HTTP DELETE request with given filename
        Waits for timeout seconds for a reply (default: 10).
        '''
        if not self.connected:
            return ""
        request = "DELETE " + resource + " HTTP/1.1\r\nHost: " + self.host + "\r\n\r\n" 
        self._debug("Sending HTTP GET request: " + request.replace('\n', '<lf>').replace('\r', '<cr>'))
        self.send(request)    
        response = ""
        first = True
        while True:
            if first:
                reply = self.getReply(timeout) 
                first = False
            else:
                reply = self.getReply(2) # decrease timeout after first reply
            if reply == None: # Timeout
                self._debug("Timeout in getReply()")
                break
            elif len(reply) == 0:  # Server closed connection
                self._debug("Server closed connection")
                break
            response += str(reply, 'utf-8')
        self._debug("Server response:\n" + response)
        self.closeConnection() # Release resources
        return response
    
    @staticmethod
    def pushover(token_key, user_key, title, message, encode = True, verbose = False):
        '''
        Sends a push request to the Pushover host api.pushover.net with given token key, user key, title and message.
        Waits for timeout seconds for a reply (default: 10).
        '''
        PUSHOVER_HOST = 'api.pushover.net'
        PUSHOVER_PORT = 443
        PUSHOVER_QUERY = '/1/messages.json'
        
        client = HTTPClient(verbose)
        client.connect(PUSHOVER_HOST, PUSHOVER_PORT)
        if not client.connected:
            return ""
        data =  'token=' + HTTPClient.toUrl(token_key, encode)
        data += '&user=' + HTTPClient.toUrl(user_key, encode)
        data += '&title=' + HTTPClient.toUrl(title, encode)
        data += '&message=' + HTTPClient.toUrl(message, encode) 
        return client.sendPostRequest(PUSHOVER_QUERY, data)

    @staticmethod    
    def toUrl(string, encode):
        if not encode:
            return string
        SAFE_CHARS = '\nabcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_.- '
        r = []
        for c in string:
            if c in SAFE_CHARS:
                r.append(c)
            else:
                r.append('%%%x' % ord(c))
        return (''.join(r)).replace(' ', '+')
    
    @staticmethod
    def extractJSON(response):
        try:
            message = response[response.find('{'):response.rfind('}') + 1]
            return ujson.loads(message)
        except:
            return None
          
    @staticmethod
    def getHeader(response):
        return response[0:response.find('\r\n\r\n')]

    @staticmethod
    def getContent(response):
        return response[response.find('\r\n\r\n') + 4:]

    @staticmethod
    def request(url, verbose = False):
        # http://host/inquiry
        try:
            url = url.strip()
            idx = url.index('://')
            protocol = url[:idx]
            idx1 = url.find('/', idx + 3)
            if idx1 == -1:
                host = url[idx + 3:] 
                query = "/"
            else:
                host = url[idx + 3:idx1] 
                query = url[idx1:]
        except:
            print("Illegal url:", url)
            return ""
        if protocol.lower() == "http":
            port = 80
        elif protocol.lower() == "https":
            port = 443
        else:
            print("Illegal url:", url)
            return ""
        client = HTTPClient(verbose)
        if verbose:
            client._debug("protocol: " + protocol + " - host: " + host + " - query: " + query)
        if not client.connect(host, port):
            print("Failed to connect to host ", host)
            return ""
        response = client.sendGetRequest(query)
        response = response[response.find('\r\n\r\n') + 4:]
        return response

     
        