# tsl2561.py
# Digital Light Sensor TSL2561
# Driver adapted from adafruit_tsl2561.py, with thanks to the author
# Version 2.01 - Jan 19, 2019

from utils import I2C_com
from machine import Pin

_COMMAND_BIT         = const(0x80)
_WORD_BIT            = const(0x20)

_CONTROL_POWERON     = const(0x03)
_CONTROL_POWEROFF    = const(0x00)

_REGISTER_CONTROL    = const(0x00)
_REGISTER_TIMING     = const(0x01)
_REGISTER_CHAN0_LOW  = const(0x0C)
_REGISTER_CHAN1_LOW  = const(0x0E)
_REGISTER_ID         = const(0x0A)

_GAIN_SCALE = (16, 1)
_TIME_SCALE = (1 / 0.034, 1 / 0.252, 1)
_CLIP_THRESHOLD = (4900, 37000, 65000)


class TSL2561():
    def __init__(self, scl = Pin(22), sda = Pin(21), addr = 0x29):
        self._i2c_com = I2C_com(addr, scl, sda)
        partno, revno = self._getChipId()
        if not partno == 5:
            raise RuntimeError('Failed to find TSL2561! Part 0x%x Rev 0x%x' % (partno, revno))
        self.setEnabled(True)

    # ---------------------- public methods ---------------------
    def getEnabled(self):
        """Returns the state of the sensor."""
        return (self._read_register(_REGISTER_CONTROL) & 0x03) != 0

    def setEnabled(self, enable):
        """Sets the state of the sensor."""
        if enable:
            self._enable()
        else:
            self._disable()

    def getLux(self):
        """Returns the computed lux value."""
        return self._compute_lux()

    def getBroadband(self):
        """Returns the broadband channel value."""
        return self._read_broadband()

    def getInfrared(self):
        """Returns the infrared channel value."""
        return self._read_infrared()

    def getLuminosity(self):
        """Returns the luminosity value. The overall luminosity as a tuple containing the broadband
        channel and the infrared channel value."""
        return (self.getBroadband(), self.getInfrared())

    def getGain(self):
        """Returns the gain (0:1x, 1:16x)."""
        return self._read_register(_REGISTER_TIMING) >> 4 & 0x01

    def setGain(self, value):
        """Sets the gain (0:1x, 1:16x)."""
        value &= 0x01
        value <<= 4
        current = self._read_register(_REGISTER_TIMING)
        cmd = _COMMAND_BIT | _REGISTER_TIMING
        v = (current & 0xef) | value
        self._i2c_com.writeReg(cmd, v)

    def getIntegrationTime(self):
        """Gets the integration time (0:13.7ms, 1:101ms, 2:402ms, or 3:manual)."""
        current = self._read_register(_REGISTER_TIMING)
        return current & 0x03

    def setIntegrationTime(self, value):
        """Sets the integration time (0:13.7ms, 1:101ms, 2:402ms, or 3:manual)."""
        value &= 0x03
        current = self._read_register(_REGISTER_TIMING)
        cmd = _COMMAND_BIT | _REGISTER_TIMING
        v = (current & 0xfc) | value
        self._i2c_com.writeReg(cmd, v)
        
    # ---------------------- private methods ---------------------
    def _compute_lux(self):
        # Based on datasheet
        ch0, ch1 = self.getLuminosity()
        if ch0 == 0:
            return None
        if ch0 > _CLIP_THRESHOLD[self.getIntegrationTime()]:
            return None
        if ch1 > _CLIP_THRESHOLD[self.getIntegrationTime()]:
            return None
        ratio = ch1 / ch0
        if ratio >= 0 and ratio <= 0.50:
            lux = 0.0304 * ch0 - 0.062 * ch0 * ratio**1.4
        elif ratio <= 0.61:
            lux = 0.0224 * ch0 - 0.031 * ch1
        elif ratio <= 0.80:
            lux = 0.0128 * ch0 - 0.0153 * ch1
        elif ratio <= 1.30:
            lux = 0.00146 * ch0 - 0.00112 * ch1
            lux = 0
        # Pretty sure the floating point math formula on pg. 23 of datasheet
        # is based on 16x gain and 402ms integration time. Need to scale
        # result for other settings.
        # Scale for gain.
        lux *= _GAIN_SCALE[self.getGain()]
        # Scale for integration time.
        lux *= _TIME_SCALE[self.getIntegrationTime()]
        return lux

    def _getChipId(self):
        chip_id = self._read_register(_REGISTER_ID)
        partno = (chip_id >> 4) & 0x0f
        revno = chip_id & 0x0f
        return (partno, revno)

    def _enable(self):
        self._write_control_register(_CONTROL_POWERON)

    def _disable(self):
        self._write_control_register(_CONTROL_POWEROFF)

    def _read_register(self, reg, count=1):
        cmd = _COMMAND_BIT | reg
        if count == 2:
            cmd |= _WORD_BIT
        data = self._i2c_com.readBlock(cmd, count)
        if count == 1:
            return data[0]
        elif count == 2:
            return data[0], data[1]
        return None

    def _write_control_register(self, data):
        cmd = _COMMAND_BIT | _REGISTER_CONTROL
        self._i2c_com.writeReg(cmd, data)

    def _read_broadband(self):
        low, high = self._read_register(_REGISTER_CHAN0_LOW, 2)
        return high << 8 | low

    def _read_infrared(self):
        low, high = self._read_register(_REGISTER_CHAN1_LOW, 2)
        return high << 8 | low

