# utils.py
# Version 1.08 - Sept 15, 2018

from machine import Pin, I2C
import utime

def epochToDateTime(epoch):
    t = utime.localtime(epoch)
    datetime = {'year' : t[0] - 30, 
                'month' : t[1],
                'day' : t[2] + 1,
                'hour' : t[3],
                'minute' : t[4],
                'second' : t[5],
                'weekday' : t[6],
                'yearday' : t[7]}
    return datetime

def getTime(epoch):
    datetime = epochToDateTime(epoch)
    hour = datetime['hour']
    min = datetime['minute']
    sec = datetime['second']
    return "%02d:%02d:%02d" %(hour, min, sec)

def getDateTime(epoch):
    datetime = epochToDateTime(epoch)
    weekday = _toDay(datetime['weekday'])
    year = datetime['year']
    month = datetime['month']
    day = datetime['day']
    hour = datetime['hour']
    min = datetime['minute']
    sec = datetime['second']
    return "%2s %4d-%02d-%02d %02d:%02d:%02d" %(weekday, year, month, day, hour, min, sec)

def _toDay(d):
    if d == 1:
        return "Su"
    if d == 2:
        return "Mo"
    if d == 3:
        return "Tu"
    if d == 4:
        return "We"
    if d == 5:
        return "Th"
    if d == 6:
        return "Fr"
    if d == 7:
        return "Sa"
    return ""
    

def toCentigrades(kelvin):
    t = "%5.1f" %(kelvin - 273.15)
    return float(t)

def cat(filename):
    with open(filename) as f:
        content = f.readlines()
    lines = [x.rstrip() for x in content]
    for line in lines:
        print(line)
        
# ------------------- Simple vector class --------------
class Vector:
    def __init__(self, *args):
        if len(args) == 2:
            self._vec = [args[0], args[1]]
        elif len(args) == 1:
            self._vec = list(args[0][:])
        else:
            self._vec = [0, 0]
            
    def add(self, otherVec):
        return Vector(self._vec[0] + otherVec[0], self._vec[1] + otherVec[1])

    def sub(self, otherVec):
        return Vector(self._vec[0] - otherVec[0], self._vec[1] - otherVec[1])

    def mult(self, k):
        return Vector(self._vec[0] * k, self._vec[1] * k)
        
    def __str__(self):
        return str(self._vec)    
    
    def __getitem__(self, index):
        return self._vec[index]

    def __setitem__(self, index, value):
         self._vec[index] = value
         
    def __eq__(self, other):
        return self._vec[0] == other._vec[0] and self._vec[1] == other._vec[1]
    
    def __ne__(self, other):
        return not self.__eq__(other)     
         
# ------------------- Class I2C_com --------------
'''
From API doc:
        i2c.init(I2C.MASTER)
        i2c.writeto(0x42, '123')        # send 3 bytes to slave with address 0x42
        i2c.writeto(addr=0x42, b'456')  # keyword for address

    Master also has other methods::

        i2c.scan()                          # scan for slaves on the bus, returning
                                            #   a list of valid addresses
        i2c.readfrom_mem(0x42, 2, 3)        # read 3 bytes from memory of slave 0x42,
                                            #   starting at address 2 in the slave
        i2c.writeto_mem(0x42, 2, 'abc')     # write 'abc' (3 bytes) to memory of slave 0x42
                                            # starting at address 2 in the slave, timeout after 1 second
'''
class I2C_com:
    i2c = None

    def __init__(self, i2c_address, scl = Pin(22), sda = Pin(21)):
        try:
            if I2C_com.i2c == None:
                I2C_com.i2c = I2C(scl = scl, sda = sda)
        except:
            raise Exception("Can't initialize I2C communication. Reset the board")        
        self.i2c_address = i2c_address

   
    def writeTo(self, data):
        '''
        Sends list/tuple of byte data (int) or bytes data without a command.
        '''
        try:
            v = bytes(bytearray(data))
            I2C_com.i2c.writeto(self.i2c_address, v)
        except:
            raise Exception("Failed to communicate with I2C device at address 0x%X: " % self.i2c_address)        
       
    def readFrom(self, count):
        '''
        Reads len number of bytes without a command.
        Returns bytes array.
        '''
        try:
            data = I2C_com.i2c.readfrom(self.i2c_address, count)
        except:
            raise Exception("Failed to communicate with I2C device at address 0x%X: " % self.i2c_address)        
        return data

    def readReg(self, cmd):
        '''
        Sends command cmd (int) and reads one byte.
        Returns byte value (int)
        '''
        try:
           data = ord(I2C_com.i2c.readfrom_mem(self.i2c_address, cmd, 1))
        except:
            raise Exception("Failed to communicate with I2C device at address 0x%X: " % self.i2c_address)        
        return data
    
    def writeReg(self, cmd, data):
        '''
        Sends commmand cmd (int) with one byte data (int).
        '''
        try:
            v = bytes(bytearray([data]))
            I2C_com.i2c.writeto_mem(self.i2c_address, cmd, v)
        except:
            raise Exception("Failed to communicate with I2C device at address 0x%X: " % self.i2c_address)        

    def readBlock(self, cmd, len):
        '''
        Sends command cmd (int) and reads len number of bytes.
        Returns list with byte values (ints)
        '''
        try:
            data = list(I2C_com.i2c.readfrom_mem(self.i2c_address, cmd, len))
        except:
            raise Exception("Failed to communicate with I2C device at address 0x%X: " % self.i2c_address)        
        return data
    
    def writeBlock(self, cmd, data):
        '''
        Sends commmand cmd (int) with bytes data as list/tuple as int (data may also be of type bytes)
        '''
        try:
            v = bytes(bytearray(data))
            I2C_com.i2c.writeto_mem(self.i2c_address, cmd, v)
        except:
            raise Exception("Failed to communicate with I2C device at address 0x%X: " % self.i2c_address)        
            
    def writeToMem(self, addr, mv):
        '''
        Write to memory using a memoryview.
        '''
        try:
            I2C_com.i2c.writeto_mem(self.i2c_address, addr, mv)
        except:
            raise Exception("Failed to communicate with I2C device at address 0x%X: " % self.i2c_address)        
        
    def readFromMem(self, addr, len):
        '''
        Reads len bytes from memory starting at addr.
        '''
        try:
            data = I2C_com.i2c.readfrom_mem(self.i2c_address, addr, len)
        except:
            raise Exception("Failed to communicate with I2C device at address 0x%X: " % self.i2c_address)        
        return data

    def readFromMemInto(self, addr, mv):
        '''
        Read from memory into memoryview.
        '''
        try:
            I2C_com.i2c.readfrom_mem_into(self.i2c_address, addr, mv)
        except:
            raise Exception("Failed to communicate with I2C device at address 0x%X: " % self.i2c_address)        

